/* c401.c: **********************************************************}
{* Tma: Rekurzivn implementace operac nad BVS
**                                         Vytvoil: Petr Pikryl, listopad 1994
**                                         pravy: Andrea Nmcov, prosinec 1995
**                                                      Petr Pikryl, duben 1996
**                                                   Petr Pikryl, listopad 1997
**                                  Pevod do jazyka C: Martin Tuek, jen 2005
**                                         pravy: Bohuslav Kena, listopad 2009
**
** Implementujte rekurzivnm zpsobem operace nad binrnm vyhledvacm
** stromem (BVS; v anglitin BST - Binary Search Tree).
**
** Klem uzlu stromu je jeden znak (obecn jm me bt cokoliv, podle
** eho se vyhledv). Uitenm (vyhledvanm) obsahem je zde integer.
** Uzly s menm klem le vlevo, uzly s vtm klem le ve stromu
** vpravo. Vyuijte dynamickho pidlovn pamti.
** Rekurzivnm zpsobem implementujte nsledujc funkce:
**
**   BSTInit ...... inicializace vyhledvacho stromu
**   BSTSearch .... vyhledvn hodnoty uzlu zadanho klem
**   BSTInsert .... vkldn nov hodnoty
**   BSTDelete .... zruen uzlu se zadanm klem
**   BSTDispose ... zruen celho stromu
**
** ADT BVS je reprezentovn koenovm ukazatelem stromu (typ tBSTNodePtr).
** Uzel stromu (struktura typu tBSTNode) obsahuje kl (typu char), podle
** kterho se ve stromu vyhledv, vlastn obsah uzlu (pro jednoduchost
** typu int) a ukazatel na lev a prav podstrom (LPtr a RPtr). 
** Pesnou definici typ naleznete v souboru c401.h.
**
** Pozor! Je teba sprvn rozliovat, kdy pout dereferenn opertor *
** (typicky pi modifikaci) a kdy budeme pracovat pouze se samotnm ukazatelem 
** (nap. pi vyhledvn). V tomto pkladu vm napov prototypy funkc.
** Pokud pracujeme s ukazatelem na ukazatel, pouijeme dereferenci.
**/

#include "c401.h"
int solved;

void BSTInit (tBSTNodePtr *RootPtr) {
/*   -------
** Funkce provede poten inicializaci stromu ped jeho prvnm pouitm.
**
** Ovit, zda byl ji strom pedan pes RootPtr inicializovn, nelze,
** protoe ped prvn inicializac m ukazatel nedefinovanou (tedy libovolnou)
** hodnotu. Programtor vyuvajc ADT BVS tedy mus zajistit, aby inicializace
** byla volna pouze jednou, a to ped vlastn prac s BVS. Proveden
** inicializace nad neprzdnm stromem by toti mohlo vst ke ztrt pstupu
** k dynamicky alokovan pamti (tzv. "memory leak").
**	
** Vimnte si, e se v hlavice objevuje typ ukazatel na ukazatel.	
** Proto je teba pi piazen pes RootPtr pout dereferenn opertor *.
** Ten bude pouit i ve funkcch BSTDelete, BSTInsert a BSTDispose.
**/
	*RootPtr = NULL;
}	

int BSTSearch (tBSTNodePtr RootPtr, char K, int *Content)	{
/*  ---------
** Funkce vyhled uzel v BVS s klem K.
**
** Pokud je takov nalezen, vrac funkce hodnotu TRUE a v promnn Content se
** vrac obsah pslunho uzlu.Pokud pslun uzel nen nalezen, vrac funkce
** hodnotu FALSE a obsah promnn Content nen definovn (nic do n proto
** nepiazujte).
**
** Pi vyhledvn v binrnm stromu bychom typicky pouili cyklus ukonen
** testem dosaen listu nebo nalezen uzlu s klem K. V tomto ppad ale
** problm ete rekurzivnm voln tto funkce, piem nedeklarujte dnou
** pomocnou funkci.
**/
	if ( RootPtr == NULL ) //strom je prazdny
		return FALSE;
	else
	{//strom neni prazdny
		if ( RootPtr->Key != K )
		{// klic K se nenasel
			if ( RootPtr->Key < K ) //klic je vetsi hledej vpravo:
				return BSTSearch(RootPtr->RPtr, K, Content);
			else //klic je mensi hledej vlevo
				return BSTSearch(RootPtr->LPtr, K, Content);
		}
		else
		{//klic se nasel
			*Content = RootPtr->BSTNodeCont;  //vracime obsah uzlu
			return TRUE;
		}
	}
} 


void BSTInsert (tBSTNodePtr* RootPtr, char K, int Content)	{	
/*   ---------
** Vlo do stromu RootPtr hodnotu Content s klem K.
**
** Pokud ji uzel se zadanm klem ve stromu existuje, bude obsah uzlu
** s klem K nahrazen novou hodnotou. Pokud bude do stromu vloen nov
** uzel, bude vloen vdy jako list stromu.
**
** Funkci implementujte rekurzivn. Nedeklarujte dnou pomocnou funkci.
**
** Rekurzivn implementace je mn efektivn, protoe se pi kadm
** rekurzivnm zanoen ukld na zsobnk obsah uzlu (zde integer).
** Nerekurzivn varianta by v tomto ppad byla efektivnj jak z hlediska
** rychlosti, tak z hlediska pamovch nrok. Zde jde ale o koln
** pklad, na kterm si chceme ukzat eleganci rekurzivnho zpisu.
**/
	if ( *RootPtr != NULL )
	{//hledame volne misto pro novy uzel
		if ( (*RootPtr)->Key > K )
			BSTInsert(&((*RootPtr)->LPtr), K, Content); //hledej vpravo
		else if ( (*RootPtr)->Key < K )
			BSTInsert(&((*RootPtr)->RPtr), K, Content); //hledej vlevo
		else
			(*RootPtr)->BSTNodeCont = Content; //z-aktualizuj obsah
	}
	else
	{//vytvarime novy uzel
		tBSTNodePtr new_ptr;
		new_ptr = malloc(sizeof(struct tBSTNode));
		if ( new_ptr == NULL ) //chybna alokace
			return;
		new_ptr->Key = K; //zapis klice K
		new_ptr->BSTNodeCont = Content; //zapiseme hodnotu Content
		//nebude mit zadne syny
		new_ptr->LPtr = NULL;
		new_ptr->RPtr = NULL;
		*RootPtr = new_ptr; //prirazeni do stromu
	}

}

void ReplaceByRightmost (tBSTNodePtr PtrReplaced, tBSTNodePtr *RootPtr) {
/*   ------------------
** Pomocn funkce pro vyhledn, pesun a uvolnn nejpravjho uzlu.
**
** Ukazatel PtrReplaced ukazuje na uzel, do kterho bude pesunuta hodnota
** nejpravjho uzlu v podstromu, kter je uren ukazatelem RootPtr.
** Pedpokld se, e hodnota ukazatele RootPtr nebude NULL (zajistte to
** testovnm ped voln tto funkce). Tuto funkci implementujte rekurzivn. 
**
** Tato pomocn funkce bude pouita dle. Ne ji zanete implementovat,
** pette si koment k funkci BSTDelete(). 
**/
	tBSTNodePtr pom_ptr = NULL; //pomocny ukazatel, bude slouzit pro smazani

	if ( *RootPtr != NULL )
	{
		if( (*RootPtr)->RPtr != NULL )
			ReplaceByRightmost(PtrReplaced, (&((*RootPtr)->RPtr))); //doprava
		else
		{
			PtrReplaced->Key = (*RootPtr)->Key;  //prepsani klice
			PtrReplaced->BSTNodeCont = (*RootPtr)->BSTNodeCont; //prepsani obsahu
			pom_ptr = *RootPtr;	//ulozeni ukazatele na koren
			*RootPtr = (*RootPtr)->LPtr; //uvolneni uzlu
			free(pom_ptr);	//odstraneni uzlu z pameti
		}
	}

}

void BSTDelete (tBSTNodePtr *RootPtr, char K) 		{
/*   ---------
** Zru uzel stromu, kter obsahuje kl K.
**
** Pokud uzel se zadanm klem neexistuje, nedl funkce nic. 
** Pokud m ruen uzel jen jeden podstrom, pak jej zdd otec ruenho uzlu.
** Pokud m ruen uzel oba podstromy, pak je ruen uzel nahrazen nejpravjm
** uzlem levho podstromu. Pozor! Nejpravj uzel nemus bt listem.
**
** Tuto funkci implementujte rekurzivn s vyuitm dve deklarovan
** pomocn funkce ReplaceByRightmost.
**/
	tBSTNodePtr pom_ptr = NULL; //pomocny ukazatel, bude slouzit pro smazani

	if ( (*RootPtr) != NULL )
	{//hledej tak dlouho dokud koren neni NULL
		if ( (*RootPtr)->Key > K )
			BSTDelete(&((*RootPtr)->LPtr), K); //hledej vpravo
		else if ( (*RootPtr)->Key < K )
			BSTDelete(&((*RootPtr)->RPtr), K); //hledej vlevo
		else
		{//uzel chtery chceme zrusit
			pom_ptr = *RootPtr; //uschovani ukazatele
			if ( pom_ptr->LPtr == NULL )
			{//overime levy podstrom
				//pom_ptr = *RootPtr; //uschovani ukazatele
				*RootPtr = pom_ptr->RPtr; //pripojeni praveho podstromu
				free(pom_ptr);
			}
			else if ( pom_ptr->RPtr == NULL )
			{//overime pravy podstrom
				//pom_ptr = *RootPtr;
				*RootPtr = pom_ptr->LPtr;
				free(pom_ptr);
			}
			else //nalezeni nejpravejsiho:
				ReplaceByRightmost(*RootPtr, (&((*RootPtr)->LPtr)));
		}
	}

} 

void BSTDispose (tBSTNodePtr *RootPtr) {	
/*   ----------
** Zru cel binrn vyhledvac strom a korektn uvoln pam.
**
** Po zruen se bude BVS nachzet ve stejnm stavu, jako se nachzel po
** inicializaci. Tuto funkci implementujte rekurzivn bez deklarovn pomocn
** funkce.
**/
	if ( *RootPtr != NULL )
	{//kde nic neni ani cert nebere
		BSTDispose(&(*RootPtr)->RPtr);	//pruchod do prava
		BSTDispose(&(*RootPtr)->LPtr);	//pruchod do leva		
		free(*RootPtr);//odstraneni uzlu z pameti
		*RootPtr = NULL;
	}
}

/* konec c401.c */

